#!/usr/bin/env python3

from __future__ import print_function

import argparse

import yaml
import psycopg2
import os
import sys

from pum.core.checker import Checker
from pum.core.dumper import Dumper
from pum.core.upgrader import Upgrader
from pum.core.exceptions import PgDumpCommandError, PgDumpFailed, PgRestoreCommandError, PgRestoreFailed
from pum.utils.utils import ask_for_confirmation, Bcolors


class Pum:
    def __init__(self, config_file=None):

        self.upgrades_table = None
        self.delta_dirs = None
        self.backup_file = None
        self.ignore_list = None
        self.pg_dump_exe = os.environ.get('PG_DUMP_EXE')
        self.pg_restore_exe = os.environ.get('PG_RESTORE_EXE')

        if config_file:
            self.__load_config_file(config_file)

    def __load_config_file(self, config_file):
        """Load the configurations from yaml configuration file and store it
        to instance variables.

        Parameters
        ----------
        config_file: string
            The path of the config file
        """
        configs = yaml.safe_load(open(config_file))
        self.set_configs(configs)

    def set_configs(self, configs):
        """Save the configuration values into the instance variables.

        Parameters
        ----------
        configs: dict
            Dictionary of configurations
        """

        self.upgrades_table = configs.get('upgrades_table')
        self.delta_dirs = configs.get('delta_dirs')
        self.backup_file = configs.get('backup_file')
        self.ignore_list = configs.get('ignore_elements')
        self.pg_dump_exe = configs.get('pg_dump_exe')
        self.pg_restore_exe = configs.get('pg_restore_exe')

        if self.delta_dirs and not isinstance(self.delta_dirs, list):
            self.delta_dirs = [self.delta_dirs]

    def run_check(self, pg_service1, pg_service2, ignore_list,
                  exclude_schema, exclude_field_pattern,
                  verbose_level=1, output_file=None):
        """Run the check command

        Parameters
        ----------
        pg_service1: string
            The name of the postgres service (defined in pg_service.conf)
            related to the first db to be compared
        pg_service2: string
            The name of the postgres service (defined in pg_service.conf)
            related to the first db to be compared
        ignore_list: list of strings
            List of elements to be ignored in check (ex. tables, columns,
            views, ...)
        exclude_schema: list of strings
            List of schemas to be ignored in check.
        exclude_field_pattern: list of strings
            List of field patterns to be ignored in check.
        verbose_level: int
            verbose level, 0 -> nothing, 1 -> print first 80 char of each
            difference, 2 -> print all the difference details
        output_file: string
            a file path where write the differences

        Returns
        -------
        True if no differences are found, False otherwise.
        """

        self.__out('Check...', type='WAITING')
        if not verbose_level:
            verbose_level = 1
        if not ignore_list:
            ignore_list = []
        if not exclude_schema:
            exclude_schema = []
        if not exclude_field_pattern:
            exclude_field_pattern = []
        try:
            checker = Checker(
                pg_service1, pg_service2,
                exclude_schema=exclude_schema, exclude_field_pattern=exclude_field_pattern,
                ignore_list=ignore_list, verbose_level=verbose_level)
            result, differences = checker.run_checks()

            if result:
                self.__out('OK', 'OKGREEN')
            else:
                self.__out('DIFFERENCES FOUND', 'WARNING')

            if differences:
                if output_file:
                    with open(output_file, 'w') as f:
                        for k, values in differences.items():
                            f.write(k)
                            f.writelines(values)
                else:
                    for k, values in differences.items():
                        print(k)
                        for v in values:
                            print(v)
            return result

        except psycopg2.Error as e:
            self.__out('ERROR', 'FAIL')
            self.__out(e.args[0], 'FAIL')
            exit(1)

        except Exception as e:
            self.__out('ERROR', 'FAIL')
            self.__out(e.args[0])
            exit(1)

    def run_dump(self, pg_service, file, exclude_schema):
        """
        Run the dump command

        Parameters
        ----------
        pg_service: string
            The name of the postgres service (defined in
            pg_service.conf) related to the first db to be compared
        file: string
            The path of the desired backup file
        """

        self.__out('Dump...', type='WAITING')

        try:
            dumper = Dumper(pg_service, file)
            if self.pg_dump_exe:
                dumper.pg_backup(pg_dump_exe=self.pg_dump_exe, exclude_schema=exclude_schema)
            else:
                dumper.pg_backup(exclude_schema=exclude_schema)
        except (PgDumpFailed, PgDumpCommandError) as e:
            self.__out('ERROR', 'FAIL')
            self.__out(e.args[0])
            exit(1)
        self.__out('OK', 'OKGREEN')

    def run_restore(self, pg_service, file, ignore_restore_errors, exclude_schema=None):
        """
        Run the dump command

        Parameters
        ----------
        pg_service: string
            The name of the postgres service (defined in
            pg_service.conf) related to the first db to be compared
        file: string
            The path of the desired backup file
        ignore_restore_errors: Boolean
            If true the pg_restore errors don't cause the exit of the program
        """

        self.__out('Restore...', type='WAITING')

        try:
            dumper = Dumper(pg_service, file)
            if self.pg_restore_exe:
                dumper.pg_restore(pg_restore_exe=self.pg_restore_exe, exclude_schema=exclude_schema)
            else:
                dumper.pg_restore(exclude_schema=exclude_schema)
        except PgRestoreFailed as e:
            self.__out('ERROR', 'FAIL')
            self.__out(str(e))
            # one might want to handle exit on error
            if ignore_restore_errors:
                return
            else:
                exit(1)
        except PgRestoreCommandError as e:
            self.__out('ERROR', 'FAIL')
            self.__out(e.args[0])
            exit(1)
        self.__out('OK', 'OKGREEN')

    def run_baseline(self, pg_service, table, delta_dirs, baseline):
        """
        Run the baseline command. Set the current database version
        (baseline) into the specified table.

        Parameters
        -----------
        pg_service: str
            The name of the postgres service (defined in
            pg_service.conf)
        table: str
            The name of the upgrades information table in the format
            schema.table
        delta_dirs: list(str)
            The paths to the delta directories
        baseline: str
            The version of the current database to set in the information
            table. The baseline must be in the format x.x.x where x are numbers.

        """

        self.__out('Set baseline...', type='WAITING')

        try:
            upgrader = Upgrader(pg_service, table, delta_dirs)
            upgrader.create_upgrades_table()
            upgrader.set_baseline(baseline)

        except ValueError as e:
            self.__out('ERROR', 'FAIL')
            self.__out(e)
            exit(1)

        self.__out('OK', 'OKGREEN')

    def run_info(self, pg_service, table, delta_dirs):
        """Print info about delta file and about already made upgrade

        Parameters
        -----------
        pg_service: str
            The name of the postgres service (defined in
            pg_service.conf)
        table: str
            The name of the upgrades information table in the format
            schema.table
        delta_dir: list(str)
            The paths to the delta directories
        """

        try:
            upgrader = Upgrader(pg_service, table, delta_dirs)
            upgrader.show_info()

        except Exception as e:
            print(e)
            exit(1)

    def run_upgrade(self, pg_service, table, delta_dirs, variables, max_version, verbose):
        """Apply the delta files to upgrade the database

        Parameters
        -----------
        pg_service: str
            The name of the postgres service (defined in pg_service.conf)
        table: str
            The name of the upgrades information table in the format
            schema.table
        delta_dirs: list(str)
            The paths to the delta directories
        variables: dict
            dictionary for variables to be used in SQL deltas ( name => value )
        max_version: str
            Maximum (including) version to run the deltas up to.
        verbose: bool
            Whether to display extra information
        """

        self.__out('Upgrade...', type='WAITING')

        try:
            upgrader = Upgrader(pg_service, table, delta_dirs, variables=variables, max_version=max_version)
            upgrader.run(verbose=verbose)

        except Exception as e:
            print(e)
            exit(1)

        self.__out('OK', 'OKGREEN')

    def run_test_and_upgrade(
            self, pg_service_prod, pg_service_test, pg_service_comp, file,
            table, delta_dirs, ignore_list, exclude_schema, exclude_field_pattern,
            ignore_restore_errors, variables, max_version, verbose):
        """
        Do the following steps:
            - creates a dump of the production db
            - restores the db dump into a test db
            - applies the delta files found in the delta directory to the test
                db.
            - checks if there are differences between the test db and a
                comparison db
            - if no significant differences are found, after confirmation,
            applies the delta files to the production dbD.


        pg_service_prod: str
            The name of the postgres service (defined in pg_service.conf)
            related to the production database
        pg_service_test:
            The name of the postgres service (defined in pg_service.conf)
            related to the test database
        pg_service_comp:
            The name of the postgres service (defined in pg_service.conf)
            related to the comparison database
        file:
            The path of the desired backup file
        table: str
            The name of the upgrades information table in the format
            schema.table
        delta_dirs: list(str)
            The paths to the delta directories
        ignore_list: list of strings
            List of elements to be ignored in check (ex. tables, columns,
            views, ...)
        exclude_schema: list of strings
            List of schemas to be ignored in check.
        exclude_field_pattern: list of strings
            List of field patterns to be ignored in check.
        ignore_restore_errors: Boolean
            If true the pg_restore errors don't cause the exit of the program
        variables: dict
            dictionary for variables to be used in SQL deltas ( name => value )
        max_version: str
            Maximum (including) version to run the deltas up to.
        verbose: bool
            Whether to display extra information

        Returns
        -------
        False if the prod database cannot be upgraded because there are
        differences between the test and comp databases.
        """

        self.__out('Test and upgrade...', type='WAITING')

        # Backup of db prod
        self.run_dump(pg_service_prod, file, exclude_schema)

        # Restore db dump on db test
        self.run_restore(pg_service_test, file, ignore_restore_errors)

        # Apply deltas on db test
        self.run_upgrade(pg_service_test, table, delta_dirs, variables, max_version, verbose)

        # Compare db test with db comp
        check_result = self.run_check(
            pg_service_test, pg_service_comp,
            ignore_list=ignore_list, exclude_schema=exclude_schema, exclude_field_pattern=exclude_field_pattern)

        if not check_result:
            return False

        if ask_for_confirmation(prompt='Apply deltas to {}?'.format(
                pg_service_prod)):
            self.run_upgrade(pg_service_prod, table, delta_dirs, variables, max_version, verbose)

        self.__out('OK', 'OKGREEN')

        return True

    def __out(self, message, type='DEFAULT'):

        # taken from django
        supported_platform = sys.platform != 'win32' or 'ANSICON' in os.environ

            # print output of the commands
        if supported_platform:
            if type == 'WAITING':
                print(Bcolors.WAITING + message + Bcolors.ENDC, end='')
            elif type == 'OKGREEN':
                print(Bcolors.OKGREEN + message + Bcolors.ENDC)
            elif type == 'WARNING':
                print(Bcolors.WARNING + message + Bcolors.ENDC)
            elif type == 'FAIL':
                print(Bcolors.FAIL + message + Bcolors.ENDC)
            elif type == 'BOLD':
                print(Bcolors.BOLD + message + Bcolors.ENDC)
            elif type == 'UNDERLINE':
                print(Bcolors.UNDERLINE + message + Bcolors.ENDC)
            else:
                print(message)
        else:
            print(message)


def create_parser() -> argparse.ArgumentParser:
    """
    Creates the main parser with its sub-parsers
    """
    # create the top-level parser
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-v", "--version", help="print the version and exit",
        action='store_true')
    parser.add_argument("-c", "--config_file", help="set the config file")

    subparsers = parser.add_subparsers(
        title='commands', description='valid pum commands', dest='command')

    # create the parser for the "check" command

    parser_check = subparsers.add_parser(
        'check', help='check the differences between two databases')
    parser_check.add_argument(
        '-p1', '--pg_service1', help='Name of the first postgres service',
        required=True)
    parser_check.add_argument(
        '-p2', '--pg_service2', help='Name of the second postgres service',
        required=True)
    parser_check.add_argument(
        '-i', '--ignore', help='Elements to be ignored', nargs='+',
        choices=['tables',
                 'columns',
                 'constraints',
                 'views',
                 'sequences',
                 'indexes',
                 'triggers',
                 'functions',
                 'rules'])
    parser_check.add_argument(
        '-N', '--exclude-schema', help='Schema to be ignored.', action='append')
    parser_check.add_argument(
        '-P', '--exclude-field-pattern', help='Fields to be ignored based on a pattern compatible with SQL LIKE.',
        action='append')
    parser_check.add_argument(
        '-v', '--verbose_level', help='Verbose level (0, 1 or 2)', type=int)
    parser_check.add_argument(
        '-o', '--output_file', help='Output file')

    # create the parser for the "dump" command
    parser_dump = subparsers.add_parser('dump', help='dump a Postgres database')

    parser_dump.add_argument(
        '-p', '--pg_service', help='Name of the postgres service',
        required=True)
    parser_dump.add_argument(
        '-N', '--exclude-schema', help='Schema to be ignored.', action='append')
    parser_dump.add_argument('file', help='The backup file')

    # create the parser for the "restore" command
    parser_restore = subparsers.add_parser(
        'restore', help='restore a Postgres database from a dump file')
    parser_restore.add_argument(
        '-p', '--pg_service', help='Name of the postgres service',
        required=True)
    parser_restore.add_argument(
        '-x', help='ignore pg_restore errors', action="store_true")
    parser_restore.add_argument(
        '-N', '--exclude-schema', help='Schema to be ignored.', action='append')
    parser_restore.add_argument('file', help='The backup file')

    # create the parser for the "baseline" command
    parser_baseline = subparsers.add_parser(
        'baseline', help='Create upgrade information table and set baseline')
    parser_baseline.add_argument(
        '-p', '--pg_service', help='Name of the postgres service',
        required=True)
    parser_baseline.add_argument(
        '-t', '--table', help='Upgrades information table', required=True)
    parser_baseline.add_argument(
        '-d', '--dir', nargs='+', help='Delta directories (space-separated)',
        required=True)
    parser_baseline.add_argument(
        '-b', '--baseline', help='Set baseline in the format x.x.x',
        required=True)

    # create the parser for the "info" command
    parser_info = subparsers.add_parser('info', help='show info about upgrades')
    parser_info.add_argument(
        '-p', '--pg_service', help='Name of the postgres service',
        required=True)
    parser_info.add_argument(
        '-t', '--table', help='Upgrades information table', required=True)
    parser_info.add_argument(
        '-d', '--dir', nargs='+', help='Delta directories (space-separated)',
        required=True)

    # create the parser for the "upgrade" command
    parser_upgrade = subparsers.add_parser('upgrade', help='upgrade db')
    parser_upgrade.add_argument(
        '-p', '--pg_service', help='Name of the postgres service',
        required=True)
    parser_upgrade.add_argument(
        '-t', '--table', help='Upgrades information table', required=True)
    parser_upgrade.add_argument(
        '-d', '--dir', nargs='+', help='Delta directories (space-separated)',
        required=True)
    parser_upgrade.add_argument(
        '-u', '--max-version', help='upper bound limit version')
    parser_upgrade.add_argument(
        '-v', '--var', nargs=3, help='Assign variable for running SQL deltas.'
                                     'Format is: (string|float|int) name value. ',
        action='append', required=False)
    parser_upgrade.add_argument(
        '-vv', '--verbose', action='store_true', help='Display extra information')

    # create the parser for the "test-and-upgrade" command
    parser_test_and_upgrade = subparsers.add_parser(
        'test-and-upgrade',
        help='try the upgrade on a test db and if all it\'s ok, do upgrade '
             'the production db')
    parser_test_and_upgrade.add_argument(
        '-pp', '--pg_service_prod',
        help='Name of the pg_service related to production db')
    parser_test_and_upgrade.add_argument(
        '-pt', '--pg_service_test',
        help='Name of the pg_service related to a test db used to test the '
             'migration')
    parser_test_and_upgrade.add_argument(
        '-pc', '--pg_service_comp',
        help='Name of the pg_service related to a db used to compare the '
             'updated db test with the last version of the db')
    parser_test_and_upgrade.add_argument(
        '-t', '--table', help='Upgrades information table')
    parser_test_and_upgrade.add_argument(
        '-d', '--dir', nargs='+', help='Delta directories (space-separated)',
        required=True)
    parser_test_and_upgrade.add_argument('-f', '--file', help='The backup file')
    parser_test_and_upgrade.add_argument(
        '-x', help='ignore pg_restore errors', action="store_true")
    parser_test_and_upgrade.add_argument(
        '-i', '--ignore', help='Elements to be ignored', nargs='+',
        choices=['tables',
                 'columns',
                 'constraints',
                 'views',
                 'sequences',
                 'indexes',
                 'triggers',
                 'functions',
                 'rules'])
    parser_test_and_upgrade.add_argument(
        '-N', '--exclude-schema', help='Schema to be ignored.', action='append')
    parser_test_and_upgrade.add_argument(
        '-P', '--exclude-field-pattern', help='Fields to be ignored based on a pattern compatible with SQL LIKE.',
        action='append')
    parser_test_and_upgrade.add_argument(
        '-u', '--max-version', help='upper bound limit version')
    parser_test_and_upgrade.add_argument(
        '-v', '--var', nargs=3, help='Assign variable for running SQL deltas.'
                                     'Format is: (string|float|int) name value.',
        action='append', required=False)
    parser_test_and_upgrade.add_argument(
        '-vv', '--verbose', action='store_true', help='Display extra information')

    return parser


def main():
    # TODO refactor and set p1 and p2 as positional args, and uniform args
    parser = create_parser()

    args = parser.parse_args()

    # print the version and exit
    if args.version:
        print('pum version {}'.format('[DEV]'))
        parser.exit()

    # if no command is passed, print the help and exit
    if not args.command:
        parser.print_help()
        parser.exit()

    variables = {}
    if args.command in ('upgrade', 'test-and-upgrade'):
        for v in args.var or ():
            if v[0] == 'float':
                variables[v[1]] = float(v[2])
            elif v[0] == 'int':
                variables[v[1]] = int(v[2])
            else:
                variables[v[1]] = v[2]

    exit_code = 0
    pum = Pum(args.config_file)

    if args.command == 'check':
        success = pum.run_check(args.pg_service1, args.pg_service2, ignore_list=args.ignore,
                                exclude_schema=args.exclude_schema, exclude_field_pattern=args.exclude_field_pattern,
                                verbose_level=args.verbose_level, output_file=args.output_file)
        if not success:
            exit_code = 1
    elif args.command == 'dump':
        pum.run_dump(args.pg_service, args.file, args.exclude_schema)
    elif args.command == 'restore':
        pum.run_restore(args.pg_service, args.file, args.x, args.exclude_schema)
    elif args.command == 'baseline':
        pum.run_baseline(args.pg_service, args.table, args.dir, args.baseline)
    elif args.command == 'info':
        pum.run_info(args.pg_service, args.table, args.dir)
    elif args.command == 'upgrade':
        pum.run_upgrade(args.pg_service, args.table, args.dir, variables, args.max_version,
                        args.verbose)
    elif args.command == 'test-and-upgrade':
        success = pum.run_test_and_upgrade(
            args.pg_service_prod, args.pg_service_test, args.pg_service_comp,
            args.file, args.table, args.dir, args.ignore,
            args.exclude_schema, args.exclude_field_pattern,
            args.x, variables, args.max_version, args.verbose)
        if not success:
            exit_code = 1

    return exit_code


if __name__ == "__main__":
    """
    Main process
    """
    _exit_code = main()
    exit(_exit_code)






import os
import shutil
import unittest
import tempfile

import psycopg2
import psycopg2.extras

from pum.core.dumper import Dumper
from pum.core.exceptions import PgRestoreFailed
from pum import __main__ as pum


class TestDumper(unittest.TestCase):
    """Test the class Dumper.

    2 pg_services needed for test related to empty db:
        pum_test_1
        pum_test_2
    """

    def tearDown(self):
        self.cur1.execute('DROP SCHEMA IF EXISTS test_dumper CASCADE;')
        self.conn1.commit()
        self.conn1.close()

        self.cur2.execute('DROP SCHEMA IF EXISTS test_dumper CASCADE;')
        self.conn2.commit()
        self.conn2.close()

        self.tmpdir.cleanup()
        self.tmp = None

    def setUp(self):
        self.pg_service1 = 'pum_test_1'
        self.pg_service2 = 'pum_test_2'

        self.conn1 = psycopg2.connect("service={0}".format(self.pg_service1))
        self.cur1 = self.conn1.cursor()

        self.conn2 = psycopg2.connect("service={0}".format(self.pg_service2))
        self.cur2 = self.conn2.cursor()

        self.cur1.execute("""
            DROP SCHEMA IF EXISTS test_dumper CASCADE;
            CREATE SCHEMA test_dumper;
            CREATE TABLE test_dumper.dumper_table
                (
                id serial NOT NULL PRIMARY KEY,
                version character varying(50),
                description character varying(200) NOT NULL,
                type integer NOT NULL
                );
            """)
        self.conn1.commit()

        self.cur2.execute("""
            DROP SCHEMA IF EXISTS test_dumper CASCADE;""")
        self.conn2.commit()

        self.tmpdir = tempfile.TemporaryDirectory()
        self.tmp = self.tmpdir.name

    def test_dump_restore(self):
        os.makedirs(self.tmp + '/test_dumper/', exist_ok=True)

        dumper = Dumper(self.pg_service1, self.tmp + '/test_dumper/dump.sql')
        dumper.pg_backup(exclude_schema=['public'])

        dumper = Dumper(self.pg_service2, self.tmp + '/test_dumper/dump.sql')
        dumper.pg_restore()

        # postgres > 9.4
        self.cur2.execute("SELECT to_regclass('{}');".format('test_dumper.dumper_table'))
        self.assertIsNotNone(self.cur2.fetchone()[0])

    def test_restore_ignore_errors(self):
        self.cur2.execute("""CREATE SCHEMA test_dumper;""")
        self.conn2.commit()

        os.makedirs(self.tmp + '/test_dumper', exist_ok=True)

        file = self.tmp + '/test_dumper/dump_ie.sql'

        dumper = Dumper(self.pg_service1, file)
        dumper.pg_backup(exclude_schema=['public'])

        dumper = Dumper(self.pg_service2, file)
        try:
            dumper.pg_restore()
            self.assertTrue(False)
        except PgRestoreFailed:
            pass

        self.cur2.execute("""DROP SCHEMA IF EXISTS test_dumper CASCADE;""")
        self.conn2.commit()
        self.cur2.execute("""CREATE SCHEMA test_dumper;""")
        self.conn2.commit()

        pum.Pum().run_restore(self.pg_service2, file, True, ['public'])

        self.cur2.execute("SELECT to_regclass('{}');".format('test_dumper.dumper_table'))
        self.assertIsNotNone(self.cur2.fetchone()[0])






if __name__ == '__main__':
    unittest.main()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from __future__ import print_function
import re
import os
from os import listdir
from os.path import isfile, join, basename, dirname
from collections import OrderedDict
import pkg_resources
import psycopg2
import psycopg2.extras
from hashlib import md5
import importlib.util
import inspect

from .deltapy import DeltaPy


class Upgrader:
    """This class is used to upgrade an existing database using sql delta files.

    Stores the info about the upgrade in a table on the database."""

    def __init__(self, pg_service, upgrades_table, dirs, variables=None,
                 max_version=None):
        """Constructor

            Parameters
            ----------
            pg_service: str
                The name of the postgres service (defined in pg_service.conf)
                related to the db
            upgrades_table: str
                The name of the table (int the format schema.name) where the
                informations about the upgrades are stored
            dirs: list(str)
                The paths to directories where delta files are stored
            variables: dict
                dictionary for variables to be used in SQL deltas ( name => value )
            max_version: str
                Maximum (including) version to run the deltas up to.
        """
        self.pg_service = pg_service
        self.connection = psycopg2.connect("service={}".format(pg_service))
        self.cursor = self.connection.cursor()
        self.upgrades_table = upgrades_table
        self.dirs = dirs
        self.variables = variables
        self.max_version = pkg_resources.parse_version(max_version) if max_version else None

    def run(self, verbose=False):
        if not self.exists_table_upgrades():
            raise UpgradesTableNotFoundError(self.upgrades_table)

        self.__run_pre_all()

        deltas = self.__get_delta_files()
        deltas = [delta for dirname in deltas for delta in deltas[dirname]]
        if not deltas:
            print('No delta files found')

        db_version = self.current_db_version()

        for d in deltas:
            if verbose:
                print('Found delta {}, version {}, type {}'.format(
                    d.get_name(), d.get_version(), d.get_type()))
                print('     Already applied: ', self.__is_applied(d))
                print('     Version greather or equal than current: ',
                      d.get_version() >= db_version)
            if not self.__is_applied(d) and d.get_version() >= db_version and \
                    (self.max_version is None or d.get_version() <= self.max_version):
                print('     Applying delta {} {}...'.format(
                    d.get_version(), d.get_name()), end=' ')

                if d.get_type() & DeltaType.SQL:
                    self.__run_delta_sql(d)
                    print('OK')
                elif d.get_type() & DeltaType.PYTHON:
                    self.__run_delta_py(d)
                    print('OK')
                else:
                    print('Delta not applied')
            else:
                if verbose:
                    print('Delta not applied')

        self.__run_post_all()

    def exists_table_upgrades(self):
        """Return if the upgrades table exists

        Returns
        -------
        bool
            True if the table exists
            False if the table don't exists"""

        query = """
            SELECT EXISTS (
            SELECT 1
            FROM   information_schema.tables
            WHERE  table_schema = '{}'
            AND    table_name = '{}'
            );
        """.format(self.upgrades_table[:self.upgrades_table.index('.')],
                   self.upgrades_table[self.upgrades_table.index('.')+1:])

        self.cursor.execute(query)
        return self.cursor.fetchone()[0]

    def __get_dbname(self):
        """Return the db name."""
        return self.connection.get_dsn_parameters()['dbname']

    def __get_dbuser(self):
        """Return the db user"""
        return self.connection.get_dsn_parameters()['user']

    def __get_delta_files(self):
        """Search for delta files and return a dict of Delta objects, keyed by directory names."""
        files = [(d, f) for d in self.dirs for f in listdir(d) if isfile(join(d, f))]

        deltas = OrderedDict()
        for d, f in files:
            file_ = join(d, f)

            if not Delta.is_valid_delta_name(file_):
                continue

            delta = Delta(file_)

            if d not in deltas:
                deltas[d] = []
            deltas[d].append(delta)

        # sort delta objects in each bucket
        for d in deltas:
            deltas[d].sort(key=lambda x: (x.get_version(), x.get_priority(), x.get_name()))

        return deltas

    def __run_delta_sql(self, delta):
        """Execute the delta sql file on the database"""

        self.__run_sql_file(delta.get_file())
        self.__update_upgrades_table(delta)

    def __run_delta_py(self, delta):
        """Execute the delta py file"""

        self.__run_py_file(delta.get_file(), delta.get_name())
        self.__update_upgrades_table(delta)

    def __run_pre_all(self):
        """Execute the pre-all.py and pre-all.sql files if they exist"""

        # if the list of delta dirs is [delta1, delta2] the pre scripts of delta2 are
        # executed before the pre scripts of delta1

        for d in reversed(self.dirs):
            pre_all_py_path = os.path.join(d, 'pre-all.py')
            if os.path.isfile(pre_all_py_path):
                print('     Applying pre-all.py...', end=' ')
                self.__run_py_file(pre_all_py_path, 'pre-all')
                print('OK')

            pre_all_sql_path = os.path.join(d, 'pre-all.sql')
            if os.path.isfile(pre_all_sql_path):
                print('     Applying pre-all.sql...', end=' ')
                self.__run_sql_file(pre_all_sql_path)
                print('OK')

    def __run_post_all(self):
        """Execute the post-all.py and post-all.sql files if they exist"""

        # if the list of delta dirs is [delta1, delta2] the post scripts of delta1 are
        # executed before the post scripts of delta2

        for d in self.dirs:
            post_all_py_path = os.path.join(d, 'post-all.py')
            if os.path.isfile(post_all_py_path):
                print('     Applying post-all.py...', end=' ')
                self.__run_py_file(post_all_py_path, 'post-all')
                print('OK')

            post_all_sql_path = os.path.join(d, 'post-all.sql')
            if os.path.isfile(post_all_sql_path):
                print('     Applying post-all.sql...', end=' ')
                self.__run_sql_file(post_all_sql_path)
                print('OK')

    def __run_sql_file(self, filepath):
        """Execute the sql file at the passed path

        Parameters
        ----------
        filepath: str
            the path of the file to execute"""

        with open(filepath, 'r') as delta_file:
            sql = delta_file.read()
            if self.variables:
                self.cursor.execute(sql, self.variables)
            else:
                self.cursor.execute(sql)
            self.connection.commit()

    def __run_py_file(self, filepath, module_name):
        """Execute the python file at the passed path

        Parameters
        ----------
        filepath: str
            the path of the file to execute
        module_name: str
            the name of the python module
            """

        # Import the module
        spec = importlib.util.spec_from_file_location(module_name, filepath)
        delta_py = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(delta_py)

        # Get the python file's directory path
        # Note: we add a separator for backward compatibility, as existing DeltaPy subclasses
        # may assume that delta_dir ends with a separator
        dir_ = dirname(filepath) + os.sep

        # Search for subclasses of DeltaPy
        for name in dir(delta_py):
            obj = getattr(delta_py, name)
            if inspect.isclass(obj) and not obj == DeltaPy and issubclass(
                    obj, DeltaPy):

                delta_py_inst = obj(
                    self.current_db_version(), dir_, self.dirs, self.pg_service,
                    self.upgrades_table, variables=self.variables)
                delta_py_inst.run()

    def show_info(self):
        """Print info about found delta files and about already made upgrades"""

        deltas = self.__get_delta_files()

        table = [['Version', 'Name', 'Type', 'Status']]

        for dir_ in deltas:
            print('delta files in dir: ', dir_)

            for delta in deltas[dir_]:
                line = [str(delta.get_version()), delta.get_name()]
                if delta.get_type() == DeltaType.PRE_PYTHON:
                    line.append('pre py')
                elif delta.get_type() == DeltaType.PRE_SQL:
                    line.append('pre sql')
                elif delta.get_type() == DeltaType.PYTHON:
                    line.append('delta py')
                elif delta.get_type() == DeltaType.SQL:
                    line.append('delta sql')
                elif delta.get_type() == DeltaType.POST_PYTHON:
                    line.append('post py')
                elif delta.get_type() == DeltaType.POST_SQL:
                    line.append('post sql')

                if self.__is_applied(delta):
                    line.append('Applied')
                else:
                    line.append('Pending')

                table.append(line)

        self.__print_table(table)

        print('')
        print('Applied upgrades in database')

        query = """SELECT
                version,
                description,
                type,
                installed_by,
                installed_on,
                success
                FROM {}
                """.format(self.upgrades_table)

        self.cursor.execute(query)
        records = self.cursor.fetchall()

        table = [['Version', 'Name', 'Type', 'Installed by', 'Installed on',
                  'Status']]

        for i in records:
            line = [str(i[0]), str(i[1])]
            delta_type = i[2]
            if delta_type == 0:
                line.append('baseline')
            elif delta_type == DeltaType.PRE_PYTHON:
                line.append('pre py')
            elif delta_type == DeltaType.PRE_SQL:
                line.append('pre sql')
            elif delta_type == DeltaType.PYTHON:
                line.append('delta py')
            elif delta_type == DeltaType.SQL:
                line.append('delta sql')
            elif delta_type == DeltaType.POST_PYTHON:
                line.append('post py')
            elif delta_type == DeltaType.POST_SQL:
                line.append('post sql')

            line.append(str(i[3]))
            line.append(str(i[4]))

            success = str(i[5])
            if success == 'True':
                line.append('Success')
            else:
                line.append('Failed')

            table.append(line)

        self.__print_table(table)

    @staticmethod
    def __print_table(table):
        """Print a list in tabular format
        Based on https://stackoverflow.com/a/8356620"""

        col_width = [max(len(x) for x in col) for col in zip(*table)]
        print("| " + " | ".join("{:{}}".format(x, col_width[i])
                                for i, x in enumerate(table[0])) + " |")
        print("| " + " | ".join("{:{}}".format('-' * col_width[i], col_width[i])
                                for i, x in enumerate(table[0])) + " |")
        for line in table[1:]:
            print("| " + " | ".join("{:{}}".format(x, col_width[i])
                                    for i, x in enumerate(line)) + " |")

    def __is_applied(self, delta):
        """Verifies if delta file is already applied on database

        Parameters
        ----------
        delta: Delta object
            The delta object representing the delta file

        Returns
        -------
        bool
            True if the delta is already applied on the db
            False otherwise
        """

        query = """
        SELECT id FROM {}
        WHERE version = '{}'
            AND checksum = '{}'
            AND success = 'TRUE'
        """.format(
            self.upgrades_table, delta.get_version(), delta.get_checksum())

        self.cursor.execute(query)
        if not self.cursor.fetchone():
            return False
        else:
            return True

    def __update_upgrades_table(self, delta):
        """Add a new record into the upgrades information table about the
        applied delta

        Parameters
        ----------
        delta: Delta
            the applied delta file"""

        query = """
        INSERT INTO {} (
            --id,
            version,
            description,
            type,
            script,
            checksum,
            installed_by,
            --installed_on,
            execution_time,
            success
        ) VALUES(
            '{}',
            '{}',
            {},
            '{}',
            '{}',
            '{}',
            1,
            TRUE
        ) """.format(
            self.upgrades_table, delta.get_version(), delta.get_name(),
            delta.get_type(), delta.get_file(), delta.get_checksum(),
            self.__get_dbuser())

        self.cursor.execute(query)
        self.connection.commit()

    def create_upgrades_table(self):
        """Create the upgrades information table"""

        query = """CREATE TABLE IF NOT EXISTS {}
                (
                id serial NOT NULL,
                version character varying(50) NOT NULL,
                description character varying(200) NOT NULL,
                type integer NOT NULL,
                script character varying(1000) NOT NULL,
                checksum character varying(32) NOT NULL,
                installed_by character varying(100) NOT NULL,
                installed_on timestamp without time zone NOT NULL DEFAULT now(),
                execution_time integer NOT NULL,
                success boolean NOT NULL,
                PRIMARY KEY (id),
                EXCLUDE (version WITH =) WHERE (type = 0)
                )
        """.format(self.upgrades_table)

        self.cursor.execute(query)
        self.connection.commit()

    def set_baseline(self, version):
        """Set the baseline into the creation information table

        version: str
            The version of the current database to set in the information
            table. The baseline must be in the format x.x.x where x are numbers.
        """
        pattern = re.compile(r"^\d+\.\d+\.\d+$")
        if not re.match(pattern, version):
            raise ValueError('Wrong version format')

        query = """
                INSERT INTO {} (
                    version,
                    description,
                    type,
                    script,
                    checksum,
                    installed_by,
                    execution_time,
                    success
                ) VALUES(
                    '{}',
                    '{}',
                    {},
                    '{}',
                    '{}',
                    '{}',
                    1,
                    TRUE
                ) """.format(self.upgrades_table, version, 'baseline', 0,
                             '', '', self.__get_dbuser())
        self.cursor.execute(query)
        self.connection.commit()

    def current_db_version(self):
        """Read the upgrades information table and return the current db
        version

        Returns
        -------
        str
            the current db version
        """

        query = """
        SELECT version from {} WHERE success = TRUE ORDER BY version DESC
        """.format(self.upgrades_table)

        self.cursor.execute(query)

        return pkg_resources.parse_version(self.cursor.fetchone()[0])


class DeltaType:
    PRE = 1
    POST = 2

    PYTHON = 4
    SQL = 8

    PRE_PYTHON = PRE | PYTHON
    PRE_SQL = PRE | SQL

    POST_PYTHON = POST | PYTHON
    POST_SQL = POST | SQL


class Delta:
    """This class represent a delta file."""

    FILENAME_PATTERN = (
        r"^(delta_)(\d+\.\d+\.\d+)(_*)(\w*)\."
        r"(pre\.sql|sql|post\.sql|pre\.py|py|post\.py)$")

    @staticmethod
    def is_valid_delta_name(file):
        """Return if a file has a valid name

        A delta file name can be:
        - pre-all.py
        - pre-all.sql
        - delta_x.x.x_ddmmyyyy.pre.py
        - delta_x.x.x_ddmmyyyy.pre.sql
        - delta_x.x.x_ddmmyyyy.py
        - delta_x.x.x_ddmmyyyy.sql
        - delta_x.x.x_ddmmyyyy.post.py
        - delta_x.x.x_ddmmyyyy.post.sql
        - post-all.py
        - post-all.sql

        where x.x.x is the version number and _ddmmyyyy is an optional
        description, usually representing the date of the delta file
        """
        filename = basename(file)
        pattern = re.compile(Delta.FILENAME_PATTERN)
        if re.match(pattern, filename):
            return True
        return False

    def __init__(self, file):
        self.file = file
        filename = basename(self.file)
        pattern = re.compile(self.FILENAME_PATTERN)
        self.match = re.match(pattern, filename)

    def __repr__(self):
        return '<file: {f} (v: {v}, n: {n}>'.format(f=self.file, v=self.get_version(), n=self.get_name())

    def get_version(self):
        """Return the version of the delta file."""
        return pkg_resources.parse_version(self.match.group(2))

    def get_name(self):
        """Return the name (description) of the delta file."""
        return self.match.group(4)

    def get_checksum(self):
        """Return the md5 checksum of the delta file."""
        with open(self.file, 'rb') as f:
            cs = md5(f.read()).hexdigest()
        return cs

    def get_type(self):
        """Return the type of the delta file.

        Returns
        -------
        type: int
        """

        ext = self.match.group(5)

        if ext == 'pre.py':
            return DeltaType.PRE_PYTHON
        elif ext == 'pre.sql':
            return DeltaType.PRE_SQL
        elif ext == 'py':
            return DeltaType.PYTHON
        elif ext == 'sql':
            return DeltaType.SQL
        elif ext == 'post.py':
            return DeltaType.POST_PYTHON
        elif ext == 'post.sql':
            return DeltaType.POST_SQL

    def get_priority(self) -> int:
        """Return the priority of the file from 1 (pre) to 3 (post)

        Returns
        -------
        type: int
        """
        dtype = self.get_type()
        if dtype & DeltaType.PRE:
            return 1
        elif dtype & DeltaType.POST:
            return 3
        else:
            return 2

    def get_file(self):
        return self.file


class UpgradesTableNotFoundError(Exception):
    """raise this when Upgrades table is not present"""



# -*- coding: utf-8 -*-

import sys
import subprocess
from distutils.version import LooseVersion

from pum.core.exceptions import PgDumpFailed, PgDumpCommandError, PgRestoreFailed, PgRestoreCommandError


class Dumper:
    """This class is used to dump and restore a Postgres database."""

    def __init__(self, pg_service, file):
        self.file = file

        self.pg_service = pg_service

    def pg_backup(self, pg_dump_exe='pg_dump', exclude_schema=None):
        """Call the pg_dump command to create a db backup

        Parameters
        ----------
        pg_dump_exe: str
            the pg_dump command path
        exclude_schema: str[]
            list of schemas to be skipped
        """

        command = [
            pg_dump_exe, '-Fc', '-f', self.file,
            'service={}'.format(self.pg_service)
            ]
        if exclude_schema:
            command.insert(-1, ' '.join("--exclude-schema={}".format(schema) for schema in exclude_schema))

        try:
            if sys.version_info[1] < 7:
                output = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            else:
                output = subprocess.run(command, capture_output=True, text=True)
            if output.returncode != 0:
                raise PgDumpFailed(output.stderr)
        except TypeError as e:
            raise PgDumpCommandError('invalid command: {}'.format(' '.join(filter(None, command))))

    def pg_restore(self, pg_restore_exe='pg_restore', exclude_schema=None):
        """Call the pg_restore command to restore a db backup

        Parameters
        ----------
        pg_restore_exe: str
            the pg_restore command path
        """

        command = [
            pg_restore_exe, '-d',
            'service={}'.format(self.pg_service),
            '--no-owner'
            ]

        if exclude_schema:
            exclude_schema_available = False
            try:
                pg_version = subprocess.check_output(['pg_restore','--version'])
                pg_version = str(pg_version).replace('\\n', '').replace("'", '').split(' ')[-1]
                exclude_schema_available = LooseVersion(pg_version) >= LooseVersion("10.0")
            except subprocess.CalledProcessError as e:
                print("*** Could not get pg_restore version:\n", e.stderr)
            if exclude_schema_available:
                command.append(' '.join("--exclude-schema={}".format(schema) for schema in exclude_schema))
        command.append(self.file)

        try:
            if sys.version_info[1] < 7:
                output = subprocess.run(command, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            else:
                output = subprocess.run(command, capture_output=True, text=True)
            if output.returncode != 0:
                raise PgRestoreFailed(output.stderr)
        except TypeError as e:
            raise PgRestoreCommandError('invalid command: {}'.format(' '.join(filter(None, command))))
